@@ -133,6 +133,14 @@ static void nchan_reserve_msg_cleanup(void *pd) {
 static void nchan_output_reserve_message_queue(ngx_http_request_t *r, nchan_msg_t *msg) {
   nchan_request_ctx_t  *ctx = ngx_http_get_module_ctx(r, ngx_nchan_module);
   ngx_http_cleanup_t   *cln;
+  
+  if(msg->storage != NCHAN_MSG_SHARED) {
+    if((msg = nchan_msg_derive_alloc(msg)) == NULL) {
+      ERR("Coudln't alloc derived msg for output_reserve_message_queue");
+      return;
+    }
+  }
+  
   if(!ctx->reserved_msg_queue) {
     if((ctx->reserved_msg_queue = ngx_palloc(r->pool, sizeof(*ctx->reserved_msg_queue))) == NULL) {
       ERR("Coudln't palloc reserved_msg_queue");
@@ -157,7 +165,7 @@ static void nchan_output_reserve_message_queue(ngx_http_request_t *r, nchan_msg_
 
 //general request-output functions and the iraq and the asian countries and dated references and the, uh, such
 
-static void nchan_flush_pending_output(ngx_http_request_t *r) {
+void nchan_flush_pending_output(ngx_http_request_t *r) {
   int                        rc;
   ngx_event_t               *wev;
   ngx_connection_t          *c;
@@ -246,7 +254,7 @@ static ngx_int_t nchan_output_filter_generic(ngx_http_request_t *r, nchan_msg_t
   c = r->connection;
   wev = c->write;
 
-  if(ctx->bcp) {
+  if(ctx && ctx->bcp) {
     nchan_bufchain_pool_refresh_files(ctx->bcp);
   }
   
@@ -265,7 +273,9 @@ static ngx_int_t nchan_output_filter_generic(ngx_http_request_t *r, nchan_msg_t
       ngx_add_timer(wev, clcf->send_timeout);
     }
     if ((ngx_handle_write_event(wev, clcf->send_lowat)) != NGX_OK) {
-      flush_all_the_reserved_things(ctx);
+      if(ctx) {
+        flush_all_the_reserved_things(ctx);
+      }
       return NGX_ERROR;
     }
     return NGX_OK;
@@ -278,7 +288,9 @@ static ngx_int_t nchan_output_filter_generic(ngx_http_request_t *r, nchan_msg_t
   }
   
   if(r->out == NULL) {
-    flush_all_the_reserved_things(ctx);
+    if(ctx) {
+      flush_all_the_reserved_things(ctx);
+    }
   }
   
   return rc;
@@ -292,32 +304,51 @@ ngx_int_t nchan_output_msg_filter(ngx_http_request_t *r, nchan_msg_t *msg, ngx_c
 }
 
 void nchan_include_access_control_if_needed(ngx_http_request_t *r, nchan_request_ctx_t *ctx) {
+  ngx_str_t          *origin;
+  ngx_str_t          *allow_origin_val;
+  static ngx_str_t    true_string = ngx_string("true");
+  nchan_loc_conf_t   *cf;
   if(!ctx) {
     ctx = ngx_http_get_module_ctx(r, ngx_nchan_module);
   }
-  nchan_loc_conf_t       *cf;
   if(!ctx) {
     return;
   }
-  if(ctx->request_origin_header.data) {
+  
+  if((origin = nchan_get_header_value_origin(r, ctx)) != NULL) {
     cf = ngx_http_get_module_loc_conf(r, ngx_nchan_module);
-    nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, &cf->allow_origin);
+    
+    if(cf->allow_credentials) {
+      nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS, &true_string);
+    }
+    allow_origin_val = nchan_get_allow_origin_value(r, cf, ctx);
+    if(allow_origin_val && allow_origin_val->len == 1 && allow_origin_val->data[0]=='*') {
+      nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, allow_origin_val);
+    }
+    else { //otherwise echo the origin
+      nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, origin);
+    }
   }
 }
 
-ngx_int_t nchan_respond_status(ngx_http_request_t *r, ngx_int_t status_code, const ngx_str_t *status_line, ngx_int_t finalize) {
+ngx_int_t nchan_respond_status(ngx_http_request_t *r, ngx_int_t status_code, const ngx_str_t *status_line, ngx_chain_t *status_body, ngx_int_t finalize) {
   ngx_int_t    rc = NGX_OK;
   r->headers_out.status=status_code;
   if(status_line!=NULL) {
     r->headers_out.status_line.len =status_line->len;
     r->headers_out.status_line.data=status_line->data;
   }
-  r->headers_out.content_length_n = 0;
-  r->header_only = 1;
+  if(status_body == NULL) {
+    r->headers_out.content_length_n = 0;
+    r->header_only = 1;
+  }
   
   nchan_include_access_control_if_needed(r, NULL);
   
   rc= ngx_http_send_header(r);
+  if(status_body) {
+    rc = ngx_http_output_filter(r, status_body);
+  }
   if(finalize) {
     nchan_http_finalize_request(r, rc);
   }
@@ -331,6 +362,26 @@ ngx_int_t nchan_respond_cstring(ngx_http_request_t *r, ngx_int_t status_code, co
   return nchan_respond_string(r, status_code, content_type, &str, finalize);
 }
 
+ngx_int_t nchan_respond_sprintf(ngx_http_request_t *r, ngx_int_t status_code, const ngx_str_t *content_type, const ngx_int_t finalize, char *fmt, ...) {
+  u_char *p;
+  ngx_str_t str;
+  va_list   args;
+    
+  str.len = 1024;
+  str.data = ngx_palloc(r->pool, 1024);
+  if(str.data == NULL) {
+    return nchan_respond_status(r, status_code, NULL, NULL, finalize);
+    return NGX_ERROR;
+  }
+
+  va_start(args, fmt);
+  p = ngx_vslprintf(str.data, str.data + str.len, fmt, args);
+  va_end(args);
+  
+  str.len = p - str.data;
+  return nchan_respond_string(r, status_code, content_type, &str, finalize);
+}
+
 ngx_int_t nchan_respond_membuf(ngx_http_request_t *r, ngx_int_t status_code, const ngx_str_t *content_type, ngx_buf_t *body, ngx_int_t finalize) {
   ngx_str_t str;
   str.len = ngx_buf_size(body);
@@ -355,7 +406,6 @@ size_t msgtag_to_strptr(nchan_msg_id_t *id, char *ch) {
   uint8_t   i;
   char     *cur;
   
-  assert(max < 255);
   static char* inactive="%i,";
   static char*  active="[%i],";
   if(max == 1) {
@@ -403,12 +453,13 @@ ngx_str_t *msgid_to_str(nchan_msg_id_t *id) {
 ngx_int_t nchan_set_msgid_http_response_headers(ngx_http_request_t *r, nchan_request_ctx_t *ctx, nchan_msg_id_t *msgid) {
   ngx_str_t                 *etag, *tmp_etag;
   nchan_loc_conf_t          *cf = ngx_http_get_module_loc_conf(r, ngx_nchan_module);
-  int8_t                     output_etag = 1, cross_origin;
+  int8_t                     output_etag = 1;
+  ngx_str_t                 *origin_header;
   
   if(!ctx) {
     ctx = ngx_http_get_module_ctx(r, ngx_nchan_module);
   }
-  cross_origin = ctx && ctx->request_origin_header.data;
+  origin_header = ctx ? nchan_get_header_value_origin(r, ctx) : NULL; 
 
   if(!cf->msg_in_etag_only) {
     //last-modified
@@ -435,7 +486,7 @@ ngx_int_t nchan_set_msgid_http_response_headers(ngx_http_request_t *r, nchan_req
     if(output_etag) {
       nchan_add_response_header(r, &NCHAN_HEADER_ETAG, etag);
     }
-    if(cross_origin) {
+    if(origin_header) {
       nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS, &NCHAN_MSG_RESPONSE_ALLOWED_HEADERS);
     }
   }
@@ -443,7 +494,7 @@ ngx_int_t nchan_set_msgid_http_response_headers(ngx_http_request_t *r, nchan_req
     if(output_etag) {
       nchan_add_response_header(r, &cf->custom_msgtag_header, etag);
     }
-    if(cross_origin) {
+    if(origin_header) {
       u_char        *cur = ngx_palloc(r->pool, 255);
       if(cur == NULL) {
         return NGX_ERROR;
@@ -462,7 +513,7 @@ ngx_int_t nchan_set_msgid_http_response_headers(ngx_http_request_t *r, nchan_req
 }
 
 ngx_int_t nchan_respond_msg(ngx_http_request_t *r, nchan_msg_t *msg, nchan_msg_id_t *msgid, ngx_int_t finalize, char **err) {
-  ngx_buf_t                 *buffer = msg->buf;
+  ngx_buf_t                 *buffer = &msg->buf;
   nchan_buf_and_chain_t     *cb;
   ngx_int_t                  rc;
   ngx_chain_t               *rchain = NULL;
@@ -491,9 +542,8 @@ ngx_int_t nchan_respond_msg(ngx_http_request_t *r, nchan_msg_t *msg, nchan_msg_i
     r->header_only = 1;
   }
 
-  if (msg->content_type.data!=NULL) {
-    r->headers_out.content_type.len=msg->content_type.len;
-    r->headers_out.content_type.data = msg->content_type.data;
+  if (msg->content_type) {
+    r->headers_out.content_type = *msg->content_type;
   }
   
   if(msgid == NULL) {
@@ -518,6 +568,7 @@ ngx_int_t nchan_respond_msg(ngx_http_request_t *r, nchan_msg_t *msg, nchan_msg_i
   
   if(rchain) {
     rc= nchan_output_filter(r, rchain);
+    if(rc != NGX_OK && err) *err="failed to write data to connection socket, probably because the connection got closed";
   }
   
   if(finalize) {
@@ -594,17 +645,17 @@ ngx_table_elt_t * nchan_add_response_header(ngx_http_request_t *r, const ngx_str
   return h;
 }
 
-ngx_int_t nchan_OPTIONS_respond(ngx_http_request_t *r, const ngx_str_t *allow_origin, const ngx_str_t *allowed_headers, const ngx_str_t *allowed_methods) {
+ngx_int_t nchan_OPTIONS_respond(ngx_http_request_t *r, const ngx_str_t *allowed_headers, const ngx_str_t *allowed_methods) {
   nchan_request_ctx_t      *ctx = ngx_http_get_module_ctx(r, ngx_nchan_module);
   
   nchan_add_response_header(r, &NCHAN_HEADER_ALLOW, allowed_methods);
   
-  if(ctx && ctx->request_origin_header.data) {
+  if(ctx && nchan_get_header_value_origin(r, ctx)) {
     //Access-Control-Allow-Origin is included later
     nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_HEADERS, allowed_headers);
     nchan_add_response_header(r, &NCHAN_HEADER_ACCESS_CONTROL_ALLOW_METHODS, allowed_methods);
   }
-  return nchan_respond_status(r, NGX_HTTP_OK, NULL, 0);
+  return nchan_respond_status(r, NGX_HTTP_OK, NULL, NULL, 0);
 }
 
 void nchan_http_finalize_request(ngx_http_request_t *r, ngx_int_t code) {
